/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.runtime;

import java.util.Date;
import java.util.HashSet;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.bonita.definition.ActivityType;
import org.ow2.bonita.definition.ClassInfo;
import org.ow2.bonita.definition.JavaHook;
import org.ow2.bonita.definition.Performer;
import org.ow2.bonita.definition.PerformerAssign;
import org.ow2.bonita.definition.RoleMapper;
import org.ow2.bonita.definition.activity.AbstractActivity;
import org.ow2.bonita.facade.def.majorElement.ParticipantDefinition.ParticipantType;
import org.ow2.bonita.facade.exception.BonitaWrapperException;
import org.ow2.bonita.facade.exception.PerformerAssignInvocationException;
import org.ow2.bonita.facade.exception.RoleMapperInvocationException;
import org.ow2.bonita.facade.runtime.ActivityBody;
import org.ow2.bonita.facade.runtime.ActivityInstance;
import org.ow2.bonita.facade.runtime.ActivityState;
import org.ow2.bonita.facade.uuid.PackageDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.facade.uuid.TaskUUID;
import org.ow2.bonita.services.Recorder;
import org.ow2.bonita.services.info.ActivityInstanceCurrentInfo;
import org.ow2.bonita.services.info.TaskCurrentInfo;
import org.ow2.bonita.util.EngineEnvTool;

/**
 * is one task instance that can be assigned to an user
 * @author Marc Blachon, Guillaume Porcher, Charles Souillard, Miguel Valdes, Pierre Vigneras
 */
public class TaskRunTime {

  private static final long serialVersionUID = 1L;
  protected long dbid;

  private static final Logger LOG = Logger.getLogger(TaskRunTime.class.getName());

  protected TaskUUID taskUUID;
  protected String description = null;
  protected String userId = null;
  protected Date create = null;
  protected Date start = null;
  protected Date end = null;
  protected Date dueDate = null;

  protected ActivityState state = ActivityState.INITIAL;
  protected ActivityState stateBeforeSuspend = null;

  protected AbstractActivity activityDef;
  protected String iterationId;
  protected String activityInstanceId;
  protected String activityId;
  protected PackageDefinitionUUID packageUUID;
  protected ProcessInstanceUUID instanceUUID;
  protected XpdlExecution xpdlExecution;

  // List of users having this role
  protected Set<String> candidates = null;


  protected TaskRunTime() {
  }

  @Override
  public boolean equals(final Object obj) {
    if (obj == null) {
      return false;
    }
    if (!(obj.getClass().equals(this.getClass()))) {
      return false;
    }
    final TaskRunTime other = (TaskRunTime)obj;
    if (other.taskUUID == null) {
      return this.taskUUID == null;
    }
    return other.taskUUID.equals(this.taskUUID);
  }

  @Override
  public int hashCode() {
    return this.taskUUID.hashCode();
  }

  public TaskRunTime(final TaskUUID taskUUID, final String nodeName,
      final AbstractActivity abstractActivity, final XpdlExecution xpdlExecution) {
    this.taskUUID = taskUUID;
    this.activityDef = abstractActivity;
    this.activityId = nodeName;
    this.xpdlExecution = xpdlExecution;
    this.instanceUUID = xpdlExecution.getXpdlInstance().getUUID();
    this.packageUUID = xpdlExecution.getProcessDefinition().getPackageDefinitionUUID();
    this.iterationId = xpdlExecution.getIterationId();
    this.activityInstanceId = xpdlExecution.getActivityInstanceId();
  }

  public void resolve(final Performer performer) {
    executeRoleMapper(performer);
    executePerformerAssign(performer);
  }

  protected void executePerformerAssign(final Performer performer) {
    final ClassInfo performerAssignInfo = performer.getPerformerAssign();
    if (performerAssignInfo != null) {
      final PerformerAssign performerAssign =
        ClassDataLoader.getInstance(PerformerAssign.class, this.packageUUID, performerAssignInfo);
      final ActivityInstance<ActivityBody> activityInst =
        EngineEnvTool.getJournalQueriers().getActivityInstance(this.instanceUUID,
            this.activityId, this.iterationId, this.activityInstanceId);
      try {
        this.userId = EngineEnvTool.getHookExecutor().executePerformerAssign(
            performerAssign, activityInst, this.candidates);
      } catch (final Exception e) {
        throw new BonitaWrapperException(
            new PerformerAssignInvocationException(performerAssignInfo.toString(), e)
        );
      }
    }
  }

  protected void executeRoleMapper(final Performer performer) {
    final ClassInfo rolemapperInfo = performer.getRoleMapper();
    if (rolemapperInfo != null) {
      final RoleMapper roleMapper = ClassDataLoader.getInstance(RoleMapper.class, this.packageUUID, rolemapperInfo);
      try {
        this.candidates = EngineEnvTool.getHookExecutor().executeRoleMapper(
            roleMapper, this.instanceUUID, performer.getParticipantId());
      } catch (final Exception e) {
        throw new BonitaWrapperException(
            new RoleMapperInvocationException(rolemapperInfo.toString(), e)
        );
      }
    }
  }

  public void ready() {
    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Ready task : " + this.taskUUID + " on activity " + this.activityId);
    }
    this.state = ActivityState.READY;

    // get performer (aka participant/swimlane)
    final Performer performer = this.activityDef.getPerformer();

    final Set<String> ar = new HashSet<String>();
    //if human set userId and candidates
    if (performer.getParticipantType().equals(ParticipantType.HUMAN)) {
      this.userId = performer.getParticipantId();
      ar.add(performer.getParticipantId());
      this.candidates = ar;
    }
    //Only ROLE and ORGANIZATIONAL_UNIT can have Role mapper (ParticipantBinding)
    if (performer.getRoleMapper() != null) {
      executeRoleMapper(performer);
    } else {
      // no role mapper: set an empty candidates list
      this.candidates = ar;
    }

    // performer assignment associated to this task
    executePerformerAssign(performer);

    EngineEnvTool.getRecorder().recordTaskReady(
        new TaskCurrentInfo(
            this.getUUID(),
            this.getUserId(),
            this.getCandidates(),
            this.getDueDate(),
            EngineEnvTool.getUserId(),
            this.getState())
    );

    EngineEnvTool.getHookExecutor().executeHooks(this.activityDef.getJavaHooks(),
        this.xpdlExecution, this.activityId, JavaHook.Type.taskOnReady);
  }

  public void start(final String userId, final boolean assignTask) {
    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Task : " + this.taskUUID + " on activity " + this.activityId + " is ready");
    }
    final Recorder recorder = EngineEnvTool.getRecorder();
    if (assignTask) {
      this.userId = userId;
      recorder.recordTaskAssigned(
          new TaskCurrentInfo(
              this.getUUID(),
              this.getUserId(),
              this.getCandidates(),
              this.getDueDate(),
              EngineEnvTool.getUserId(),
              this.getState())
      );
    }
    this.state = ActivityState.EXECUTING;
    final ActivityInstanceCurrentInfo activityInstanceCurrentInfo = new ActivityInstanceCurrentInfo(
        this.xpdlExecution.getXpdlInstance().getPackageDefinitionUUID(),
        this.xpdlExecution.getXpdlInstance().getProcessDefinitionUUID(),
        this.xpdlExecution.getXpdlInstance().getUUID(),
        this.xpdlExecution.getCurrentActivityInstanceUUID(),
        ActivityType.task,
        this.getActivityId(),
        this.xpdlExecution.getIterationId(),
        this.xpdlExecution.getActivityInstanceId(),
        this.xpdlExecution.getScopeVariables());

    recorder.recordBodyStarted(activityInstanceCurrentInfo);
    recorder.recordTaskStarted(
        new TaskCurrentInfo(
            this.getUUID(),
            this.getUserId(),
            this.getCandidates(),
            this.getDueDate(),
            EngineEnvTool.getUserId(),
            this.getState())
    );

    EngineEnvTool.getHookExecutor().executeHooks(this.activityDef.getJavaHooks(),
        this.xpdlExecution, this.activityId, JavaHook.Type.taskOnStart);
  }

  public void finish(final String userId, final boolean assignTask) {
    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Finishing task : " + this.taskUUID + " on activity " + this.activityId);
    }
    final Recorder recorder = EngineEnvTool.getRecorder();
    if (assignTask) {
      this.userId = userId;
      recorder.recordTaskAssigned(
          new TaskCurrentInfo(
              this.getUUID(),
              this.getUserId(),
              this.getCandidates(),
              this.getDueDate(),
              EngineEnvTool.getUserId(),
              this.getState())
      );
    }
    this.state = ActivityState.FINISHED;
    recorder.recordTaskFinished(
        new TaskCurrentInfo(
            this.getUUID(),
            this.getUserId(),
            this.getCandidates(),
            this.getDueDate(),
            EngineEnvTool.getUserId(),
            this.getState())
    );
    EngineEnvTool.getHookExecutor().executeHooks(this.activityDef.getJavaHooks(),
        this.xpdlExecution, this.activityId, JavaHook.Type.taskOnFinish);
    this.activityDef.signal(this.xpdlExecution, AbstractActivity.BODY_FINISHED, null);
    this.xpdlExecution = null;
  }

  public void abort() {
    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Aborting task : " + this.taskUUID + " on activity " + this.activityId);
    }
    this.state = ActivityState.ABORTED;
    this.xpdlExecution = null;
  }

  public void cancel() {
    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Cancelling task : " + this.taskUUID + " on activity " + this.activityId);
    }
    this.state = ActivityState.CANCELLED;
    this.xpdlExecution = null;
  }

  public void suspend(final String userId, final boolean assignTask) {
    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Suspending task : " + this.taskUUID + " on activity " + this.activityId);
    }
    final Recorder recorder = EngineEnvTool.getRecorder();
    if (assignTask) {
      this.userId = userId;
      recorder.recordTaskAssigned(
          new TaskCurrentInfo(
              this.getUUID(),
              this.getUserId(),
              this.getCandidates(),
              this.getDueDate(),
              EngineEnvTool.getUserId(),
              this.getState())
      );
    }
    this.stateBeforeSuspend = this.state;
    this.state = ActivityState.SUSPENDED;
    recorder.recordTaskSuspended(
        new TaskCurrentInfo(
            this.getUUID(),
            this.getUserId(),
            this.getCandidates(),
            this.getDueDate(),
            EngineEnvTool.getUserId(),
            this.getState())
    );
    EngineEnvTool.getHookExecutor().executeHooks(this.activityDef.getJavaHooks(),
        this.xpdlExecution, this.activityId, JavaHook.Type.taskOnSuspend);
  }

  public void resume(final String resumeUserId, final boolean taskAssign) {
    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Resuming task : " + this.taskUUID + " on activity " + this.activityId);
    }
    final Recorder recorder = EngineEnvTool.getRecorder();
    if (taskAssign) {
      this.userId = resumeUserId;
      recorder.recordTaskAssigned(new TaskCurrentInfo(
          this.getUUID(),
          this.getUserId(),
          this.getCandidates(),
          this.getDueDate(),
          EngineEnvTool.getUserId(),
          this.getState()
      )
      );
    }
    this.state = this.stateBeforeSuspend;
    recorder.recordTaskResumed(
        new TaskCurrentInfo(
            this.getUUID(),
            this.getUserId(),
            this.getCandidates(),
            this.getDueDate(),
            EngineEnvTool.getUserId(),
            this.getState())
    );
    EngineEnvTool.getHookExecutor().executeHooks(this.activityDef.getJavaHooks(),
        this.xpdlExecution, this.activityId, JavaHook.Type.taskOnResume);
  }

  public void assign(final java.util.Set<java.lang.String> candidates, final String assignUserId) {
    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("assigning task : " + this.taskUUID + " on activity " + this.activityId);
    }
    final Recorder recorder = EngineEnvTool.getRecorder();
    this.userId = assignUserId;
    this.candidates = candidates;
    recorder.recordTaskAssigned(
        new TaskCurrentInfo(
            this.getUUID(),
            this.getUserId(),
            this.getCandidates(),
            this.getDueDate(),
            EngineEnvTool.getUserId(),
            this.getState())
    );
  }

  public String getDescription() {
    return this.description;
  }

  public void setDescription(final String description) {
    this.description = description;
  }

  public String getUserId() {
    return this.userId;
  }

  public Date getCreate() {
    return this.create;
  }

  public void setCreate(final Date create) {
    this.create = create;
  }

  public Date getDueDate() {
    return this.dueDate;
  }

  public void setDueDate(final Date dueDate) {
    this.dueDate = dueDate;
  }

  public Date getEnd() {
    return this.end;
  }

  public void setEnd(final Date end) {
    this.end = end;
  }
  public Date getStart() {
    return this.start;
  }
  public void setStart(final Date start) {
    this.start = start;
  }
  public TaskUUID getUUID() {
    return this.taskUUID;
  }
  public Set<String> getCandidates() {
    return this.candidates;
  }
  public void setCandidates(final Set<String> candidates) {
    this.candidates = candidates;
  }
  public ActivityState getState() {
    return this.state;
  }
  public void setState(final ActivityState state) {
    this.state = state;
  }

  @Override
  public String toString() {
    return "TaskRuntime : userId = " + this.userId
      + ", candidates = " + this.candidates + ", activityId = " + this.activityId;
  }


  public String getActivityId() {
    return this.activityId;
  }

  public ProcessInstanceUUID getInstanceUUID() {
    return this.instanceUUID;
  }

  public long getDbid() {
    return this.dbid;
  }


  public XpdlExecution getXpdlExecution() {
    return this.xpdlExecution;
  }


  public AbstractActivity getActivityDef() {
    return this.activityDef;
  }


  public PackageDefinitionUUID getPackageUUID() {
    return this.packageUUID;
  }

}
